import pathlib
import glob
from warnings import warn
from random import choices
import subprocess
import zipfile
import concurrent.futures
import os
from os.path import join
import shlex
import time
from collections import defaultdict
from ..utils import unzip, read_json
from ..sizes import file_size
from tqdm.auto import tqdm

DATA_PATH = join(os.path.dirname(__file__), 'data/')
__all__ = ['split_videos', 'VideoDataset']


def split_videos(meta_file):
    '''
    Groups real-fake videos in dictionary
    '''
    split = defaultdict(lambda: set())
    for vid in meta_file:
        if meta_file[vid]['label'] == 'FAKE':
            split[meta_file[vid]['original']].add(vid)
    return split


class VideoDataset:
    '''
    create dataset from videos and metadata.
    @params:

    To download and create use VideoDataset.from_part method
    '''

    def __init__(self, path, metadata_path=None):
        self.path = pathlib.Path(path)
        self.video_paths = list(self.path.glob('*.mp4'))

        metadata_path = metadata_path if metadata_path else self.path / 'metadata.json'
        try:
            self.metadata = read_json(metadata_path)
        except FileNotFoundError:
            del metadata_path
            print('metadata file not found.\n Some functionalities may not work.')

        if hasattr(self, 'metadata'):
            self.video_groups = split_videos(self.metadata)

    @staticmethod
    def download_part(part='00', download_path='.', cookies_path=join(DATA_PATH, 'kaggle', 'cookies.txt')):
        dataset_path = f'https://www.kaggle.com/c/16880/datadownload/dfdc_train_part_{part}.zip'
        folder = f'dfdc_train_part_{int(part)}'
        command = f'wget -c --load-cookies {cookies_path} {dataset_path} -P {download_path}'
        command_args = shlex.split(command)
        fp = open(os.devnull, 'w')
        download = subprocess.Popen(command_args, stdout=fp, stderr=fp)
        bar = tqdm(total=10240, desc='Downloading ')
        zip_size = 0
        while download.poll() is None:
            time.sleep(0.1)
            try:
                new_size = int(
                    file_size(download_path + f'/dfdc_train_part_{part}.zip'))
                bar.update(new_size - zip_size)
                zip_size = new_size
            except FileNotFoundError:
                continue
        if download.poll() != 0:
            print('some error')
            print('download', download.poll())
        download.terminate()
        fp.close()
        bar.close()
        return download_path + f'/dfdc_train_part_{part}.zip'

    @classmethod
    def from_part(cls, part='00',
                  cookies_path=join(DATA_PATH, 'kaggle', 'cookies.txt'),
                  download_path='.'):
        folder = f'dfdc_train_part_{int(part)}'

        if os.path.exists(pathlib.Path(download_path) / folder):
            return cls(pathlib.Path(download_path) / folder)
        downloaded_zip = cls.download_part(
            part=part, download_path=download_path, cookies_path=cookies_path)
        unzip(downloaded_zip)
        os.remove(download_path + f'/dfdc_train_part_{part}.zip')
        path = pathlib.Path(download_path) / folder
        return cls(path)

    def __len__(self):
        return len(self.video_paths)

    def n_groups(self, n, k=-1):
        '''
        returns random n real-fake pairs by default.
        else starting from k.
        '''
        if k != -1:
            if n + k >= len(self.video_groups):
                warn(RuntimeWarning(
                    'n+k is greater then video length. Returning available'))
                n = len(self.video_groups) - k - 1
            return self.video_groups[k:n + k]
        if n >= len(self.video_groups):
            warn(RuntimeWarning('n is greater then total groups. Returning available'))
            n = len(self.video_groups) - 1
        return choices(self.video_groups, k=n)



from sys import getsizeof
import os

__all__ = ['dir_size','file_size','var_size']

def byte_to_mb(size):
    return size/(1024**2)


def dir_size(start_path='.'):
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return byte_to_mb(total_size)


def file_size(path):
    file_stats = os.stat(path)
    return byte_to_mb(file_stats.st_size)


def var_size(var):
    return byte_to_mb(getsizeof(var))

__all__ = ['var_size','file_size','dir_size']


# from torch_lr_finder import LRFinder
from tqdm.auto import tqdm
from functools import partial
import torch
import time

tqdm_nl = partial(tqdm, leave=False)


def _limit_string(string, length):
    string = str(string)
    if length > len(string):
        return string
    else:
        return string[:length - 2] + '..'


def _time_rep(seconds):
    if seconds >= 3600:
        return time.strftime('%H:%M:%S', time.gmtime(seconds))
    else:
        return time.strftime('%M:%S', time.gmtime(seconds))


class Learner:
    def __init__(self, model):
        self.model = model

    def compile(self, optimizer, loss, lr_scheduler=None, device='cpu', metrics=None):
        self.optimizer = optimizer
        self.loss = loss
        self.device = device
        if lr_scheduler is not None:
            self.lr_scheduler = lr_scheduler
        if metrics is not None:
            self.metrics = metrics
        else:
            self.metrics = []

    def fit(self, epochs, train_loader, val_loader=None, accumulation_steps=1):
        # TODO: test for model on same device
        each_train_info = []
        each_val_info = []
        header_string = ''
        headings = ['Train Loss', 'Val Loss']
        for i in range(len(self.metrics)):
            headings.append(self.metrics[i].__name__)
            if i == 2:
                break

        for heading in headings:
            header_string += _limit_string(heading, 12).center(12) + '|'
        header_string += 'Time'.center(12) + '|'
        print(header_string)

        # train
        self.optimizer.zero_grad()
        for epoch in tqdm_nl(range(epochs)):
            self.model.train()
            train_info = {}
            val_info = {}
            train_info['losses'] = []

            start_time = time.perf_counter()
            train_length = len(train_loader)

            for i, (inputs, targets) in tqdm_nl(enumerate(train_loader), total=train_length, desc='Training: '):
                inputs, targets = inputs.to(
                    self.device), targets.to(self.device)
                pred = self.model(inputs)
                loss = self.loss(pred, targets)
                train_info['losses'].append(loss)
                loss.backward()

                if (i + 1) % accumulation_steps == 0:
                    # self.optimizer.step()
                    # if hasattr(self, 'lr_scheduler'):
                    #     self.lr_scheduler.step()
                    self.optimizer.zero_grad()
            train_info['time'] = time.perf_counter() - start_time

            if val_loader is not None:
                val_info = self.validate(val_loader)
            info_string = ''

            def format_infos(x, length):
                return _limit_string(round(torch.stack(x).mean().item(), 2), 12).center(12)
            info_values = [format_infos(train_info['losses'], 12)]

            if 'losses' in val_info:
                info_values.append(format_infos(val_info['losses'], 12))
            else:
                info_values.append(str(None).center(12))

            for i, metric in enumerate(self.metrics):
                info_values.append(format_infos(
                    val_info['metrics'][metric.__name__], 12))
                if i == 2:
                    break
            total_time = train_info['time']
            if 'time' in val_info:
                total_time += val_info['time']
            info_values.append(_time_rep(total_time).center(12))
            # print('|'.join(info_values), end='|\n')
            tqdm.write('|'.join(info_values) + '|')
            each_train_info.append(train_info)
            each_val_info.append(val_info)
        return {'train': each_train_info, 'val': each_val_info}

    def validate(self, val_loader):
        information = {}
        information['losses'] = []
        information['metrics'] = {}
        for metric in self.metrics:
            information['metrics'][metric.__name__] = []

        model.eval()
        val_loss = torch.zeros(1)
        start_time = time.perf_counter()
        with torch.set_grad_enabled(False):
            for inputs, targets in tqdm_nl(val_loader, desc='Validating: '):
                inputs, targets = inputs.to(
                    self.device), targets.to(self.device)
                pred = self.model(inputs)
                loss = self.loss(pred, targets)
                information['losses'].append(loss)
                for metric in self.metrics:
                    information['metrics'][metric.__name__].append(
                        metric(pred, targets))

        total_time = time.perf_counter() - start_time
        information['time'] = total_time
        return information



import json
import zipfile
from tqdm.auto import tqdm

__all__ = ['read_json','write_json','unzip']

def read_json(path):
    '''
    Read Json file as dict.
    '''
    with open(path, 'rb') as file:
        json_dict = json.load(file)
    return json_dict


def write_json(dictionary, path):
    """
    Write dict as a json file
    """
    with open(path, 'w') as fp:
        json.dump(dictionary, fp)


def unzip(zip_path, path='.'):
    with zipfile.ZipFile(zip_path) as zf:
        for member in tqdm(zf.infolist(), desc='Extracting '):
            try:
                zf.extract(member, path)
            except zipfile.error as e:
                pass
